import datetime
import functools
from http import HTTPStatus

from typing import Optional, List

from sermonaudio import API, APIException, _session, models
from sermonaudio.models import Sermon, SermonEventType, SermonSeries
from sermonaudio.utils import update_kwargs_for_key, join_url_path


class BroadcasterAPIError(APIException):  # pragma: no cover
    pass


def parse_sermon(response) -> Sermon:  # pragma: no cover
    if response.ok:
        return Sermon.parse(response.json())
    else:
        raise BroadcasterAPIError(response.json())


def check_response_code_or_throw(response) -> bool:  # pragma: no cover
    """Returns true if response is successful from an endpoint that returns nothing."""
    if response.ok:
        return True
    else:
        if response.json():
            raise BroadcasterAPIError(response.json())
        else:
            # If no error message was returned, use the status code.
            raise BroadcasterAPIError({"errors": {"status_code": response.status_code}})


class Broadcaster(API):  # pragma: no cover. Unfortunately we can't test this w/o a test account...
    @classmethod
    def create_or_update_sermon(
        cls,
        sermon_id: Optional[str],
        accept_copyright: bool,
        full_title: str,
        speaker_name: str,
        preach_date: datetime.date,
        publish_timestamp: Optional[datetime.datetime],
        event_type: SermonEventType,
        display_title: Optional[str],
        subtitle: Optional[str],
        bible_text: Optional[str],  # TODO: structure this data
        more_info_text: Optional[str],
        language_code: str,
        keywords: Optional[List[str]],
        **kwargs,
    ) -> Sermon:
        """Creates a new sermon record or updates an existing one (media uploaded separately).

        :param sermon_id: The sermon ID, if you are updating an existing sermon. Otherwise, None.
        :param accept_copyright: A boolean value indicating that you agree that you are allowed to upload this content.
        :param full_title: The full sermon title.
        :param speaker_name: The speaker name (please be consistent; speakers will be created if they don't exist).
        :param preach_date: The date the sermon was preached.
        :param publish_timestamp: The time that the sermon should be visible on the site. You must submit this in order
        for the sermon to be visible to the public. Callers are encouraged to use timezone aware objects to ensure
        that the correct timestamp is generated.
        :param event_type: The type of event that this sermon was preached at.
        :param display_title: An alternate, shorter version of the title to be displayed in ID3v1 tags and parts of
        the legacy site (30 char limit)
        :param subtitle: The subtitle of the sermon (or the series, if multiple sermons use the same subtitle; this
        will likely be reworked in the future)
        :param bible_text: The scripture passage(s) that this sermon was derived from
        :param more_info_text: Additional info about the sermon that you wish to share
        :param language_code: The ISO 639 language code for the sermon.
        :param keywords: A list of keywords for this sermon.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: The created sermon
        """
        update_kwargs_for_key(
            kwargs,
            "params",
            {
                "acceptCopyright": accept_copyright,
                "fullTitle": full_title,
                "speakerName": speaker_name,
                "preachDate": preach_date.isoformat(),
                "publishTimestamp": int(publish_timestamp.timestamp()) if publish_timestamp else None,
                "eventType": event_type.value,
                "displayTitle": display_title,
                "subtitle": subtitle,
                "bibleText": bible_text,
                "moreInfoText": more_info_text,
                "languageCode": language_code,
                "newsInFocus": "false",
                "keywords": " ".join(keywords) if keywords else None,
            },
        )

        if sermon_id:
            path = join_url_path("node", "sermons", sermon_id)
            method = "put"
        else:
            path = join_url_path("node", "sermons")
            method = "post"

        return cls._request(path, method, json=kwargs["params"], parse_func=parse_sermon, **kwargs)

    @classmethod
    def publish_sermon(cls, sermon_id: str):  # pragma: no cover
        path = join_url_path("node", "sermons", sermon_id)
        params = {"publishNow": True}
        return cls.patch(path, params)

    @classmethod
    def move_sermon_to_series(cls, sermon_id: str, series_id: str):  # pragma: no cover
        path = join_url_path("node", "sermons", sermon_id)
        params = {"series_id": series_id}
        return cls.patch(path, params)

    @classmethod
    def delete_sermon(cls, sermon_id: str):  # pragma: no cover
        path = join_url_path("node", "sermons", sermon_id)
        return cls.delete(path)

    @classmethod
    def duplicate_sermon(cls, sermon_id: str, **kwargs) -> Sermon:  # pragma: no cover
        path = join_url_path("node", "sermons", sermon_id, "duplicate")
        return cls.post(path, parse_func=parse_sermon, **kwargs)

    @classmethod
    def _upload_media(cls, upload_type: str, sermon_id: str, path: str, **kwargs):  # pragma: no cover
        """Uploads media for a sermon.
        :param upload_type: The type of media to upload.
        :param sermon_id: The Sermon ID you are uploading media for.
        :param path: The path to the file on disk.
        :param kwargs: Additional arguments to pass to the underlying post method.
        :return: No return value. Throws a BroadcasterAPIError with details on failure.
        """
        update_kwargs_for_key(kwargs, "params", {"uploadType": upload_type, "sermonID": sermon_id})

        response = cls.post("media", **kwargs)

        if response.status_code != HTTPStatus.CREATED:
            raise BroadcasterAPIError({"error": "Unable to create media upload", "json": response.json()})

        upload_url = response.json()["uploadURL"]

        with open(path, "rb") as fp:
            response = _session.post(upload_url, data=fp, stream=True, **kwargs)
            if not response.ok:
                raise BroadcasterAPIError(
                    {
                        "error": f"Received unexpected HTTP status code {response.status_code} when uploading data.",
                        "response": response.content,
                    }
                )

    upload_audio = functools.partialmethod(_upload_media, "original-audio")
    upload_video = functools.partialmethod(_upload_media, "original-video")

    #
    # Series
    #
    @classmethod
    def create_series(cls, title: str, broadcaster_id: str) -> Optional[SermonSeries]:  # pragma: no cover
        """Creates an empty series.

        The new series can't already exist.

        :param title: The title for the new series.
        :param broadcaster_id: The ID of the broadcaster that will own the series.
        :return: The series, if it was created successfully.
        """
        path = join_url_path("node", "broadcasters", broadcaster_id, "series")
        return cls.post(
            path, data={"series_name": title}, parse_func=lambda res: SermonSeries(res.json()) if res.ok else None
        )

    @classmethod
    def rename_series(cls, id_or_current_title: str, new_title: str, broadcaster_id: str) -> bool:  # pragma: no cover
        """Renames a series.

        If the new series name already exists, all sermons from the
        original series will be merged in.

        :param id_or_current_title: The ID or current title of the series.
        :param new_title: The new title for the series.
        :param broadcaster_id: The ID of the broadcaster that owns the series.
        :return: Boolean indicating success.
        """
        path = join_url_path("node", "broadcasters", broadcaster_id, "series", id_or_current_title)
        return cls.patch(path, data={"new_series_name": new_title}, parse_func=check_response_code_or_throw)

    @classmethod
    def delete_series(cls, id_or_title: str, broadcaster_id: str) -> bool:  # pragma: no cover
        """Deletes a series.

        Deleting a series does not delete the sermons in the series.

        :param id_or_title: The ID or title of the series to delete.
        :param broadcaster_id: The ID of the broadcaster that owns the series.
        :return:  Boolean indicating success.
        """
        path = join_url_path("node", "broadcasters", broadcaster_id, "series", id_or_title)
        return cls.delete(path, parse_func=check_response_code_or_throw)

    #
    # Welcome video
    #
    @classmethod
    def set_welcome_video(cls, sermon_id: Optional[str], broadcaster_id: str) -> bool:  # pragma: no cover
        """Sets the welcome video for a broadcaster.

        The welcome video is the video you see on your SermonAudio home
        page, and typically has information about your church or
        organization. The sermon_id of this video is available in the
        welcome_video field of the Broadcaster object.

        :param sermon_id: The sermon to use as the welcome video,
        which must have video available. Clear the setting by passing None.
        :param broadcaster_id: The ID of the broadcaster that owns the sermon.
        :return:  Boolean indicating success.
        """
        path = join_url_path("node", "broadcasters", broadcaster_id)
        return cls.patch(path, json={"welcomeVideoID": sermon_id}, parse_func=check_response_code_or_throw)

    #
    # Highlighted Sermons
    #
    @classmethod
    def set_highlighted_audio_sermon(
        cls,
        broadcaster_id: str,
        audio_type: models.HighlightedAudioInputTypes,
        sermon_id: Optional[str] = None,
        event_type: Optional[models.SermonEventType] = None,
    ) -> bool:  # pragma: no cover
        """Sets the highlighted audio sermon for a broadcaster.

        Each broadcaster can highlight a sermon on their home page for
        visitors to see first. The sermon can be a specific sermon_id,
        the most recent sermon in a specific event type, or one of the
        other HighlightedAudioInputTypes values.

        See the HighlightedAudioInputTypes enum for more documentation
        on the audio types you can set. If you use SERMON_ID or
        EVENT_TYPE, you must supply a corresponding sermon_id or
        event_type parameter.

        :param broadcaster_id: The ID of the broadcaster to update.
        :param audio_type: The type of audio highlighted sermon.
        :param sermon_id: Must be supplied with audio_type SERMON_ID.
        :param event_type: Must be supplied with audio_type EVENT_TYPE.
        :return:  Boolean indicating success.
        """
        if sermon_id is not None and event_type is not None:
            raise BroadcasterAPIError("Supply only one of sermon_id or event_type.")

        if audio_type is models.HighlightedAudioInputTypes.SERMON_ID and sermon_id is None:
            raise BroadcasterAPIError("No sermon_id supplied.")

        if audio_type is models.HighlightedAudioInputTypes.EVENT_TYPE and event_type is None:
            raise BroadcasterAPIError("No event_type supplied.")

        path = join_url_path("node", "broadcasters", broadcaster_id, "highlighted_sermons")

        params = {"audioType": audio_type.value}

        if sermon_id:
            params["audioSermonID"] = sermon_id

        if event_type:
            params["audioEventType"] = event_type.value

        return cls.patch(path, data=params, parse_func=check_response_code_or_throw)

    @classmethod
    def set_highlighted_sort_order(
        cls,
        broadcaster_id: str,
        sort_order: models.HighlightedSortOrders,
    ) -> bool:  # pragma: no cover
        """Sets a broadcaster's highlighted sort order.

        Broadcasters have a short list of sermons on their home page,
        and this sort order controls how they are sorted. See the
        HighlightedSortOrders enum for more details.

        :param broadcaster_id: The ID of the broadcaster to update.
        :param sort_order: The new sort order.
        :return: Boolean indicating success.
        """
        path = join_url_path("node", "broadcasters", broadcaster_id, "highlighted_sermons")
        return cls.patch(path, data={"sortOrder": sort_order.value}, parse_func=check_response_code_or_throw)



"""API data structures"""
import datetime
import enum
import pytz

_model_override = {}

_generic_speaker_names = {"Various Speakers", "Unknown Speaker"}


class MultipleOverrideError(Exception):
    pass


class Model:
    """Base model class.

    This is responsible for some of the magic surrounding model overrides. You really should
    inherit from this.
    """

    def __init__(self, obj: dict):
        super().__init__()

        self.__obj = obj

    @classmethod
    def parse(cls, obj: dict):
        return _model_override.get(cls, cls)(obj)


def override_model(model):
    def wrapper(new_model):
        if model in _model_override:
            raise MultipleOverrideError(
                f"Multiple overrides for {model}. This is almost certainly an error."
                "If you ABSOLUTELY have a good reason for doing this, you should"
                "instead override the existing subclass of this model and let the"
                "resolution chain work deterministically. Any other use is assuredly"
                "going to land you in a world of pain."
            )

        _model_override[model] = new_model

        return new_model

    return wrapper


def isinstance_or_none(obj, t):
    return obj is None or isinstance(obj, t)


class Node(Model):
    """The base node object, which encapsulates all node API responses"""

    def __init__(self, obj: dict):
        super().__init__(obj)

        self.node_type = obj["nodeType"]
        assert isinstance(self.node_type, str)

        self.node_display_name = obj["nodeDisplayName"]
        assert isinstance(self.node_display_name, str)

        self.results = obj["results"]
        assert isinstance(self.results, list) or isinstance(self.results, dict) or self.results is None

        self.total_count = obj["totalCount"]
        assert isinstance_or_none(self.total_count, int)

        self.next = obj["next"]
        assert isinstance_or_none(self.next, str)


class Broadcaster(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.broadcaster_id = obj["broadcasterID"]
        assert isinstance(self.broadcaster_id, str)

        self.id_code = obj["idCode"]
        assert isinstance(self.id_code, str)

        self.service_times_are_preformatted = obj["serviceTimesArePreformatted"]
        assert isinstance_or_none(self.service_times_are_preformatted, bool)

        self.service_times = obj["serviceTimes"]
        assert isinstance_or_none(self.service_times, str)

        self.denomination = obj["denomination"]
        assert isinstance_or_none(self.denomination, str)

        self.address = obj["address"]
        assert isinstance_or_none(self.address, str)

        self.display_name = obj["displayName"]
        assert isinstance(self.display_name, str)

        self.short_name = obj["shortName"]
        assert isinstance(self.short_name, str)

        self.location = obj["location"]
        assert isinstance(self.location, str)

        self.latitude = obj["latitude"]
        assert isinstance_or_none(self.latitude, float)

        self.longitude = obj["longitude"]
        assert isinstance_or_none(self.longitude, float)

        self.image_url = obj["imageURL"]
        assert isinstance(self.image_url, str)

        self.album_art_url_format = obj["albumArtURL"]
        assert isinstance(self.album_art_url_format, str)

        self.minister = obj["minister"]
        assert isinstance_or_none(self.minister, str)

        self.phone = obj["phone"]
        assert isinstance_or_none(self.phone, str)

        self.listen_line_number = obj["listenLineNumber"]
        assert isinstance_or_none(self.listen_line_number, str)

        self.home_page_url = obj["homePageURL"]
        assert isinstance_or_none(self.home_page_url, str)

        self.bible_version = obj["bibleVersion"]
        assert isinstance_or_none(self.bible_version, str)

        self.facebook_username = obj["facebookUsername"]
        assert isinstance_or_none(self.facebook_username, str)

        self.twitter_username = obj["twitterUsername"]
        assert isinstance_or_none(self.twitter_username, str)

        self.about_us = obj["aboutUs"]
        assert isinstance_or_none(self.about_us, str)

        self.can_webcast = obj["canWebcast"]
        assert isinstance(self.can_webcast, bool)

        self.webcast_in_progress = obj["webcastInProgress"]
        assert isinstance(self.webcast_in_progress, bool)

        self.vacant_pulpit = obj["vacantPulpit"]
        assert isinstance(self.vacant_pulpit, bool)

        self.welcome_video_id = obj["welcomeVideoID"]
        assert isinstance_or_none(self.welcome_video_id, str)

        self.country_iso_code = obj["countryISOCode"]
        assert isinstance_or_none(self.country_iso_code, str)

        self.language_code = obj["languageCode"]
        assert isinstance_or_none(self.language_code, str)

    def get_album_art_url(self, size: int):  # pragma: no cover
        """Returns a URL for the square album art with a with and height equal to the provided size argument"""
        return self.album_art_url_format.replace("{size}", str(size))

    def __str__(self):  # pragma: no cover
        return f'<Broadcaster {self.broadcaster_id} "{self.display_name}">'


class Speaker(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.display_name = obj["displayName"]
        assert isinstance(self.display_name, str)

        self.sort_name = obj.get("sortName")
        assert isinstance(self.sort_name, str)

        self.bio = obj["bio"]
        assert isinstance_or_none(self.bio, str)

        self.portrait_url = obj["portraitURL"]
        assert isinstance(self.portrait_url, str)

        self.rounded_thumbnail_image_url = obj["roundedThumbnailImageURL"]
        assert isinstance(self.rounded_thumbnail_image_url, str)

        self.album_art_url_format = obj["albumArtURL"]
        assert isinstance(self.album_art_url_format, str)

        # The following are present on certain endpoints, such as filter_options
        if obj.get("mostRecentPreachDate"):
            self.most_recent_preach_date = datetime.datetime.strptime(obj["mostRecentPreachDate"], "%Y-%m-%d").date()
        else:
            self.most_recent_preach_date = None

        self.sermon_count = obj.get("sermonCount")
        assert isinstance_or_none(self.sermon_count, int)

    def get_album_art_url(self, size: int):  # pragma: no cover
        """Returns a URL for the square album art with a with and height equal to the provided size argument"""
        return self.album_art_url_format.replace("{size}", str(size))

    @property
    def is_generic(self) -> bool:
        return self.display_name in _generic_speaker_names


@enum.unique
class MediaClass(enum.Enum):
    AUDIO = "audio"
    VIDEO = "video"
    TEXT = "text"

    ALL = "all"


@enum.unique
class MediaType(enum.Enum):
    MP3 = "mp3"
    AAC = "aac"

    MP4 = "mp4"

    PDF = "pdf"
    WORD = "doc"
    TRANSCRIPT = "transcript"

    JPEG = "jpg"

    ORIGINAL_AUDIO = "orig-audio"
    ORIGINAL_VIDEO = "orig-video"


class Media(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        try:
            self.media_type = MediaType(obj["mediaType"])
        except ValueError:
            self.media_type = None

        self.is_live = obj["live"]
        assert isinstance(self.is_live, bool)

        self.is_adaptive = obj["adaptiveBitrate"]
        assert isinstance(self.is_adaptive, bool)

        self.stream_url = obj["streamURL"]
        assert isinstance_or_none(self.stream_url, str)

        self.event_stream_url = obj["eventStreamURL"]
        assert isinstance_or_none(self.event_stream_url, str)

        self.download_url = obj.get("downloadURL")
        assert isinstance_or_none(self.download_url, str)

        self.bitrate = obj["bitrate"]
        assert isinstance_or_none(self.bitrate, int)

        try:
            self.file_size_bytes = obj["fileSizeBytes"]
        except ValueError:
            self.file_size_bytes = None

        self.duration = obj["duration"]
        assert isinstance_or_none(self.duration, int)

        self.audio_codec = obj["audioCodec"]
        assert isinstance_or_none(self.audio_codec, str)

        self.video_codec = obj["videoCodec"]
        assert isinstance_or_none(self.video_codec, str)

        self.thumbnail_image_url = obj["thumbnailImageURL"]
        assert isinstance_or_none(self.thumbnail_image_url, str)

        self.raw_url = obj.get("rawURL")
        assert isinstance_or_none(self.raw_url, str)


class MediaSet(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        def is_valid(rec: Media):  # Silently ignore parse errors
            return rec.media_type is not None

        self.audio = list(filter(is_valid, [Media.parse(rec) for rec in obj["audio"]]))
        self.video = list(filter(is_valid, [Media.parse(rec) for rec in obj["video"]]))
        self.text = list(filter(is_valid, [Media.parse(rec) for rec in obj["text"]]))


class Sermon(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.sermon_id = obj["sermonID"]
        assert isinstance(self.sermon_id, str)

        self.broadcaster = Broadcaster.parse(obj["broadcaster"])
        self.speaker = Speaker.parse(obj["speaker"])

        self.full_title = obj["fullTitle"]
        assert isinstance(self.full_title, str)

        self.display_title = obj["displayTitle"]
        assert isinstance(self.display_title, str)

        self.subtitle = obj["subtitle"]
        assert isinstance_or_none(self.subtitle, str)

        self.series = SermonSeries.parse(obj["series"]) if obj.get("series") else None

        self.preach_date = datetime.datetime.strptime(obj["preachDate"], "%Y-%m-%d").date()
        assert isinstance(self.preach_date, datetime.date)

        # A sermon may be picked as a staff pick by SermonAudio.
        try:
            self.staff_pick_date = datetime.datetime.strptime(obj["pickDate"], "%Y-%m-%d").date()
            assert isinstance_or_none(self.staff_pick_date, datetime.date)
        except (TypeError, ValueError):
            self.staff_pick_date = None

        # A sermon may be featured by a broadcaster by going to a sermon's page
        # and selecting the option to feature it. This given the sermon a high
        # level of visibility at SermonAudio. The last feature date is reported
        # by the API, since a sermon can be featured more than once.
        try:
            self.last_feature_date = datetime.datetime.strptime(obj["lastFeatureDate"], "%Y-%m-%d").date()
            assert isinstance_or_none(self.last_feature_date, datetime.date)
        except (TypeError, ValueError):
            self.last_feature_date = None

        timestamp = obj["publishTimestamp"]
        self.publish_timestamp = datetime.datetime.fromtimestamp(timestamp, pytz.utc) if timestamp is not None else None

        update_date = obj["updateDate"]
        self.update_date = datetime.datetime.fromtimestamp(update_date, pytz.utc)

        self.language_code = obj["languageCode"]
        assert isinstance(self.language_code, str)

        self.bible_text = obj["bibleText"]
        assert isinstance_or_none(self.bible_text, str)

        self.more_info_text = obj["moreInfoText"]
        assert isinstance_or_none(self.more_info_text, str)

        try:
            self.event_type = SermonEventType(obj["eventType"])
        except ValueError:
            self.event_type = None

        # Localized event type as a display string.
        self.display_event_type = obj["displayEventType"]
        assert isinstance_or_none(self.display_event_type, str)

        self.download_count = obj["downloadCount"]  # audio download count
        assert isinstance(self.download_count, int)

        self.video_download_count = obj["videoDownloadCount"]
        assert isinstance(self.video_download_count, int)

        self.document_download_count = obj["documentDownloadCount"]
        assert isinstance(self.document_download_count, int)

        self.external_link = obj["externalLink"]
        assert isinstance_or_none(self.external_link, str)

        self.keywords = obj.get("keywords", [])

        self.media = MediaSet.parse(obj["media"])

        self.archived_webcast_id = obj["archivedWebcastID"]
        assert isinstance_or_none(self.archived_webcast_id, int)

    def __str__(self):  # pragma: no cover
        return f"<Speaker {self.speaker.display_name} - {self.display_title}>"


class Webcast(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.display_name = obj["displayName"]
        assert isinstance_or_none(self.display_name, str)

        self.webcast_id = obj["webcastID"]
        assert isinstance(self.webcast_id, int)

        self.broadcaster_id = obj["broadcasterID"]
        assert isinstance(self.broadcaster_id, str)

        self.source_location = obj["broadcasterLocation"]
        assert isinstance_or_none(self.source_location, str)

        self.start_time = datetime.datetime.fromtimestamp(obj["startTime"], tz=pytz.utc)
        assert isinstance(self.start_time, datetime.datetime)

        try:
            self.end_time = datetime.datetime.fromtimestamp(obj["endTime"], tz=pytz.utc)
            assert isinstance_or_none(self.start_time, datetime.datetime)
        except TypeError:
            # live webcasts do not have an end time
            self.end_time = None

        self.preview_image_url = obj["previewImageURL"]
        assert isinstance_or_none(self.preview_image_url, str)

        self.resizable_preview_image_url = obj["resizablePreviewImageURL"]
        assert isinstance_or_none(self.resizable_preview_image_url, str)

        self.peak_listener_count = obj["peakListenerCount"]
        assert isinstance(self.peak_listener_count, int)

        self.total_tune_in_count = obj["totalTuneInCount"]
        assert isinstance(self.total_tune_in_count, int)

        self.total_listen_line_count = obj["totalListenLineCount"]
        assert isinstance(self.total_listen_line_count, int)

        self.media = MediaSet.parse(obj["media"])

        # The field archived_sermon_id has been deprecated. Originally, one
        # webcast could be archived to one sermon. By popular demand, we
        # relaxed this requirement so one webcast could be archived to multiple
        # sermons. If your webcast is linked to one sermon, this field will
        # have the sermon ID of that sermon for backwards
        # compatibility. Otherwise, it will be None. We recommend that you use
        # the archived_sermons field, which is a list of sermon IDs that have
        # been created from this webcast.
        self.archived_sermon_id = obj["archivedSermonID"]
        assert isinstance_or_none(self.archived_sermon_id, str)

        self.archived_sermons = obj["archivedSermons"]
        assert isinstance_or_none(self.archived_sermons, list)

    def __str__(self):  # pragma: no cover
        return f'<WebcastInfo {self.broadcaster_id} "{self.display_name}"">'


class RelativeBroadcasterLocation(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.broadcaster = Broadcaster.parse(obj["broadcaster"])
        assert isinstance(self.broadcaster, Broadcaster)

        self.meters = obj["meters"]
        assert isinstance(self.meters, int)

    def __str__(self):  # pragma: no cover
        return f"<RelativeBroadcasterLocation {self.broadcaster.broadcaster_id} - ~{round(self.meters/1000, 1)}km away>"


class SermonSeries(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.series_id = obj["seriesID"]
        self.title = obj["title"]
        self.broadcaster_id = obj["broadcasterID"]
        self.latest = datetime.datetime.strptime(obj["latest"][:10], "%Y-%m-%d") if obj["latest"] else None
        self.earliest = datetime.datetime.strptime(obj["earliest"][:10], "%Y-%m-%d") if obj["earliest"] else None
        self.updated = datetime.datetime.fromtimestamp(obj["updated"], pytz.utc)
        self.count = obj["count"]

    def __str__(self):  # pragma: no cover
        return f"<SermonSeries {self.title} ({self.count} sermons)>"


class SermonEventType(enum.Enum):
    AUDIO_BOOK = "Audio Book"
    BIBLE_STUDY = "Bible Study"
    CAMP_MEETING = "Camp Meeting"
    CHAPEL_SERVICE = "Chapel Service"
    CHILDREN = "Children"
    CLASSIC_AUDIO = "Classic Audio"
    CONFERENCE = "Conference"
    CURRENT_EVENTS = "Current Events"
    DEBATE = "Debate"
    DEVOTIONAL = "Devotional"
    FUNERAL_SERVICE = "Funeral Service"
    MIDWEEK_SERVICE = "Midweek Service"
    PODCAST = "Podcast"
    PRAYER_MEETING = "Prayer Meeting"
    Q_AND_A = "Question & Answer"
    RADIO_BROADCAST = "Radio Broadcast"
    SERMON_CLIP = "Sermon Clip"
    SPECIAL_MEETING = "Special Meeting"
    SUNDAY_AFTERNOON = "Sunday Afternoon"
    SUNDAY_AM = "Sunday - AM"
    SUNDAY_PM = "Sunday - PM"
    SUNDAY_SCHOOL = "Sunday School"
    SUNDAY_SERVICE = "Sunday Service"
    TEACHING = "Teaching"
    TESTIMONY = "Testimony"
    TV_BROADCAST = "TV Broadcast"
    VIDEO_DVD = "Video DVD"
    WEDDING = "Wedding"
    YOUTH = "Youth"


class SermonEventTypeDetail(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.type = obj.get("type")
        self.description = obj.get("description")
        self.display_event_type = obj.get("displayEventType")
        self.number_of_sermons = obj.get("numberOfSermons")
        self.roku_image_url = obj.get("rokuImageURL")
        self.fire_tv_image_url = obj.get("fireTVImageURL")
        self.number_of_sermons = obj.get("numberOfSermons")


@enum.unique
class SpurgeonDevotionalType(enum.Enum):
    AM = "AM"  # Morning Devotional
    PM = "PM"  # Evening Devotional
    CHECKBOOK = "CHECKBOOK"  # Faith's Checkbook (note American spelling)


class SpurgeonDevotional(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        self.type = SpurgeonDevotionalType(obj["type"])
        self.month = obj["month"]
        self.day = obj["day"]
        self.quote = obj["quote"]
        self.reference = obj["reference"]
        self.content = obj["content"]
        self.audio = Sermon.parse(obj["sermon"])

    def __str__(self):  # pragma: no cover
        return f"<SpurgeonDevotional for {self.month}/{self.day}>"


@enum.unique
class SermonSortOption(enum.Enum):
    DOWNLOADS = "downloads"
    EVENT = "event"
    LANGUAGE = "language"
    LAST_PLAYED = "lastplayed"
    NEWEST = "newest"
    OLDEST = "oldest"
    PICK_DATE = "pickdate"
    SERIES = "series"
    SPEAKER = "speaker"
    UPDATED = "updated"
    RANDOM = "random"
    ADDED = "added"
    TITLE = "title"


class Book(Model):
    """A book of the bible, and associated chapters that you can use to filter a broadcaster's content"""

    def __init__(self, obj: dict):
        super().__init__(obj)

        self.name = obj["bookName"]
        self.display_name = obj["displayBookName"]
        self.chapters = obj["chapters"]
        self.osis_pa = obj["osisPA"]


class Language(Model):
    """A language that you can use to filter a broadcaster's content"""

    def __init__(self, obj: dict):
        super().__init__(obj)

        self.name = obj["languageName"]
        self.iso_code = obj["languageCode"]


class SermonCountForSpeaker(Model):
    """A speaker name, and how many sermons they preached at a broadcaster (used for filtering)"""

    def __init__(self, obj: dict):
        super().__init__(obj)

        self.count = obj["count"]
        self.speaker_name = obj["speaker"]


class FilterOptions(Model):
    """A set of various things that you can filter by.

    For example, which books and chapters of the Bible have content, which speakers
    have spoken at a given church/ministry, etc.
    """

    def __init__(self, obj: dict):
        super().__init__(obj)

        self.books = [Book.parse(rec) for rec in obj["bibleBooks"]]
        self.languages = [Language.parse(rec) for rec in obj["languages"]]
        self.series = [SermonSeries.parse(rec) for rec in obj["series"]]
        self.speakers = [Speaker.parse(rec) for rec in obj["speakers"]]
        self.sermon_counts_for_speakers = [SermonCountForSpeaker.parse(rec) for rec in obj["sermonCountsForSpeakers"]]
        self.sermon_event_types = [SermonEventTypeDetail.parse(rec) for rec in obj["sermonEventTypes"]]
        self.years: [int] = [rec["year"] for rec in obj["years"]]  # This structure is simple enough to reduce


@enum.unique
class SeriesSortOrder(enum.Enum):
    # The last date that the series was "touched" (never goes back in time)
    LAST_UPDATED = "last_updated"

    # The *create* date of the latest sermon in the series (may go back in time)
    NEWEST_SERMON_CREATE_DATE = "newest_sermon_create_date"

    # Title sort ascending
    TITLE = "title"

    # Sermon count DESC
    SERMON_COUNT_HIGHEST = "sermon_count_highest"

    # Sermon count DESC
    SERMON_COUNT_LOWEST = "sermon_count_lowest"


@enum.unique
class HighlightedAudioInputTypes(enum.Enum):
    """Input values to select the type of highlighted audio.

    You may set the highlighted sermon to a specific sermon, or to the
    most recent sermon from a specific event type. Otherwise, you may
    select the most recent sermon a broadcaster has uploaded, or a
    random sermon from the broadcaster's archive. Random and newest pick
    are reserved for future use, but not currently used.
    """

    RANDOM_SERMON = "randomsource"
    NEWEST_SERMON = "newest"
    RANDOM_PICK = "randompick"
    NEWEST_PICK = "newestpick"
    SERMON_ID = "sermon_id"
    EVENT_TYPE = "event_type"


class HighlightedSortOrders(enum.Enum):
    """Input values to select the order of highlighted sermons.

    Broadcasters have a short list sermons on their home pages, and this
    setting selects the order in which the list is sorted. As with
    HighlightedAudioInputTypes, random and newest picks are reserved but
    not currenly used.

    Note that the use of "pick" here as an enum value is retained for
    historical compatibility. These picks are listener recommendations
    ("listener picks"), not staff picks.
    """

    PREACHED = "preached"
    POPULAR = "popular"
    ADDED = "added"
    NEWEST_PICKS = "picklistnewest"
    RANDOM_PICKS = "picklistrandom"
    NEWEST_CLIPS = "clipsnewest"
    RANDOM_CLIPS = "clipsrandom"
    DEFAULT = "added"

    @property
    def sermon_parameters(self):
        """Get dict of sermon parameters for current sort order.

        The sort order value, by itself, is not particularly useful, so
        this method translates it into a dict of parameters you can pass
        to Node.get_sermons() to retrieve sermons sorted in that order.

        You can directly pass a HighlightedSortOrders value as the
        sort_by parameter to Node.get_sermons().

        For more control over the Node.get_sermons() parameters, you may
        retrieve this property as a dict, update it as necessary, and
        pass it to Node.get_sermons().

        Example:
            highlighted_sermons  = Node.get_highlighted_sermons()
            params = highlighted_sermons.sort_order.sermon_parameters
            # ... update params...
            sermons = Node.get_sermons(**params)
        """

        if self is self.ADDED:
            return {"sort_by": SermonSortOption.ADDED}
        elif self is self.NEWEST_PICKS:
            return {"listener_recommended": True, "sort_by": SermonSortOption.NEWEST}
        elif self is self.RANDOM_PICKS:
            return {"listener_recommended": True, "sort_by": SermonSortOption.RANDOM}
        elif self is self.NEWEST_CLIPS:
            return {"event_type": SermonEventType.SERMON_CLIP, "sort_by": SermonSortOption.NEWEST}
        elif self is self.RANDOM_CLIPS:
            return {"event_type": SermonEventType.SERMON_CLIP, "sort_by": SermonSortOption.RANDOM}
        elif self is self.POPULAR:
            return {"sort_by": SermonSortOption.DOWNLOADS}
        else:
            return {"sort_by": SermonSortOption.NEWEST}


class HighlightedSermons(Model):
    def __init__(self, obj: dict):
        super().__init__(obj)

        if obj["sortOrder"]:
            self.sort_order = HighlightedSortOrders(obj["sortOrder"])
            assert isinstance(self.sort_order, HighlightedSortOrders)
        else:
            self.sort_order = None

        self.sort_title = obj["sortTitle"]
        assert isinstance_or_none(self.sort_title, str)

        self.audio_title = obj["audioTitle"]
        assert isinstance_or_none(self.audio_title, str)

        if obj["audioSermon"]:
            self.audio_sermon = Sermon.parse(obj["audioSermon"])
            assert isinstance(self.audio_sermon, Sermon)
        else:
            self.audio_sermon = None


@enum.unique
class ListSortOrder(enum.Enum):
    DATE = "date"
    REVERSE_DATE = "reverse-date"


class WebcastStats(Model):
    def __init__(self, obj: dict):

        # The webcast_id can be None because this class is used for cumulative
        # stats, where no specific webcast_id exists, as well as individual
        # webcasts.
        self.webcast_id = obj["webcastID"]
        assert isinstance_or_none(self.webcast_id, int)

        self.total_countries = obj["totalCountries"]
        assert isinstance(self.total_countries, int)

        self.locations = obj["locations"]
        assert isinstance(self.locations, dict)


class WebcastRangeTotals(Model):
    def __init__(self, obj: dict):
        self.broadcaster_id = obj["broadcasterID"]

        self.start_time = datetime.datetime.fromtimestamp(obj["startTime"], tz=pytz.utc)
        assert isinstance(self.start_time, datetime.datetime)

        self.end_time = datetime.datetime.fromtimestamp(obj["endTime"], tz=pytz.utc)
        assert isinstance(self.end_time, datetime.datetime)

        self.max_peak_listener_count = obj["maxPeakListenerCount"]
        assert isinstance(self.max_peak_listener_count, int)

        self.avg_total_tune_in_count = obj["averageTotalTuneInCount"]
        assert isinstance(self.avg_total_tune_in_count, int)

        self.total_webcast_views = obj["totalWebcastViews"]
        assert isinstance(self.total_webcast_views, int)

        self.total_listen_line_count = obj["totalListenLineCount"]
        assert isinstance(self.total_listen_line_count, int)

        self.cumulative_stats = WebcastStats.parse(obj["cumulativeStats"])



import functools

from typing import List, Optional, Union

from osis_book_tools import OSISBook

from requests import Response

from sermonaudio import API, APIException, models
from sermonaudio.utils import update_kwargs_for_key, join_url_path

URL_PATH = "node"


class NodeAPIError(APIException):
    pass


def parse_node(response) -> models.Node:
    if response.ok:
        return models.Node.parse(response.json())
    else:
        raise NodeAPIError(response.json())


class Node(API):
    """An interface to the SermonAudio.com node API v2.

    This class provides a variety of low-level and high-level helpers to make it
    easy to interact with the API. A number of higher-level helpers are exposed to
    make common API calls easy. However, the API has many options, and we do not
    necessarily comprehensively model them all here. Please refer to the documentation,
    and if you need to build additional methods for your application, you may use the lower
    level helpers at the top of the class to build your own `Node` subclass with additional
    high-level helper methods.
    """

    #
    # Low-level building blocks
    #

    get_node = functools.partialmethod(API.get, parse_func=parse_node)
    get_sermon_list_node = functools.partialmethod(
        get_node, parse_func=lambda res: [models.Sermon.parse(rec) for rec in parse_node(res).results]
    )

    class PaginatedResponse:
        def __init__(self, page: int, page_size: int, response: models.Node, result_processing_func=None):
            self.page = page
            self.page_size = page_size
            self.next_url = response.next
            self.total_count = response.total_count
            self.results = (
                [result_processing_func(result) for result in response.results]
                if result_processing_func
                else response.results
            )

    #
    # Sermons
    #

    @classmethod
    def get_sermons(
        cls,
        path: str = None,
        page: int = 1,
        page_size: Optional[int] = None,
        book: Optional[OSISBook] = None,
        chapter: Optional[int] = None,
        chapter_end: Optional[int] = None,
        verse: Optional[int] = None,
        verse_end: Optional[int] = None,
        event_type: Optional[models.SermonEventType] = None,
        search_keyword: Optional[str] = None,
        language_code: Optional[str] = None,
        require_audio: bool = False,
        require_video: bool = False,
        series: Optional[str] = None,
        broadcaster_id: str = None,
        sermon_ids: [str] = None,
        speaker_name: str = None,
        year: Optional[int] = None,
        sort_by: Optional[Union[models.HighlightedSortOrders, models.SermonSortOption]] = None,
        staff_pick: bool = False,
        featured: bool = False,
        listener_recommended: bool = False,
        updated: bool = False,
        include_drafts: bool = False,
        include_scheduled: bool = False,
        include_published: bool = True,
        **kwargs,
    ) -> PaginatedResponse:  # noqa: F821
        """Returns a set of sermons with pagination data, when available.

        :param path: The path to the endpoint you want to call (defaults to the all-purpose sermon node endpoint).
        :param page: The page number to load (defaults to 1).
        :param page_size: The number of items per page (currently defaults to 50 if omitted).
        :param kwargs: Additional arguments to pass to the underlying get method.
        :param book: The book of the Bible to limit results to.
        :param chapter: The chapter of the book to limit results to.
        :param chapter_end: If you want to query a range of chapters, the range will span from chapter to chapter_end.
        :param verse: The verse of the chapter to limit results to.
        :param verse_end: If you want to query a range of verses, the range will span from verse to verse_end.
        :param event_type: The type of event to limit results to, if any.
        :param search_keyword: Keywords to search by.
        :param language_code: The language code to limit results to.
        :param require_audio: If true, all results will have ready-to-play audio.
        :param require_video: If true, all results will have ready-to-play video.
        :param series: The series to limit results to (may be an ID or name).
        :param broadcaster_id: The ID of the broadcaster to limit results to.
        :param sermon_ids: A set of sermon IDs to return. This option ignores most others.
        :param speaker_name: The speaker name to limit results to.
        :param year: The year to limit results to (preach date, not upload date).
        :param sort_by: A sort to apply to the results. Defaults to preach date descending.
        :param staff_pick: List only sermons that are SermonAudio staff picks.
        :param featured: List only sermons that were featured by the broadcaster.
        :param listener_recommended: List only sermons that have been recommended by listeners.
        :param include_drafts: If true, include sermons with no publish date.
        :param include_scheduled: If true, include sermons set to be published in the future.
        :param include_published: If true, include published sermons.
        :param highlighted_sermon_sort: If included, use the sort order determined by the
        value of this parameter. If a highlighted sermon sort is included, it overrides
        other sort/filter options.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A sermon list response with pagination data, if available.
        """
        path = path or join_url_path(URL_PATH, "sermons")

        if isinstance(sort_by, models.HighlightedSortOrders):
            params = sort_by.sermon_parameters

            # Set the sort_by the the default sort order for highlighted sermons.
            sort_by = models.SermonSortOption.NEWEST

            # Update our parameters with any parameters that sermon_parameters
            # returns.
            if "listener_recommended" in params:
                listener_recommended = params["listener_recommended"]
            if "sort_by" in params:
                sort_by = params["sort_by"]
            if "event_type" in params:
                event_type = params["event_type"]

        update_kwargs_for_key(
            kwargs,
            "params",
            {
                "page": page,
                "pageSize": page_size,
                "book": book.paratext_abbreviation if book else None,
                "chapter": chapter,
                "chapterEnd": chapter_end,
                "verse": verse,
                "verseEnd": verse_end,
                "eventType": event_type.value if event_type else None,
                "searchKeyword": search_keyword,
                "languageCode": language_code,
                "requireAudio": require_audio,
                "requireVideo": require_video,
                "series": series,
                "broadcasterID": broadcaster_id,
                "sermonIDs": ",".join(sermon_ids) if sermon_ids else None,
                "speakerName": speaker_name,
                "year": year,
                "sortBy": sort_by.value if sort_by else None,
                "staffPick": staff_pick,
                "featured": featured,
                "listenerRecommended": listener_recommended,
                "includeDrafts": include_drafts,
                "includeScheduled": include_scheduled,
                "includePublished": include_published,
            },
        )

        return cls.get_node(
            path,
            parse_func=lambda res: cls.PaginatedResponse(page, page_size, parse_node(res), models.Sermon.parse),
            **kwargs,
        )

    @classmethod
    def get_sermon(cls, sermon_id: str, **kwargs) -> Optional[models.Sermon]:
        """Fetches info for a single sermon, if it exists.

        :param sermon_id: The ID of the sermon you want to fetch.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: models.Sermon or None
        """
        return cls.get(
            join_url_path(URL_PATH, "sermons", str(sermon_id)),
            parse_func=lambda res: models.Sermon.parse(res.json()) if res.ok else None,
            **kwargs,
        )

    @classmethod
    def delete_media(cls, sermon_id: str, media_type: models.MediaClass, **kwargs) -> Response:  # pragma: no cover
        """Deletes media of a given type for a single sermon.

        :param sermon_id: The ID of the sermon whose media you want to delete.
        :param media_type: The type of media you want to delete.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: bool indicating success or failure
        """
        params = {"sermonID": sermon_id, "target": media_type.value}
        return cls.post(join_url_path("jobs", "delete"), params=params, **kwargs)

    #
    # Broadcasters
    #

    @classmethod
    def get_broadcaster(cls, broadcaster_id: str, **kwargs) -> Optional[models.Broadcaster]:
        """Fetches info for a single broadcaster, if it exists.

        :param broadcaster_id: The ID of the broadcaster you want to fetch.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: models.Broadcaster or None
        """
        return cls.get(
            join_url_path(URL_PATH, "broadcasters", str(broadcaster_id)),
            parse_func=lambda res: models.Broadcaster.parse(res.json()) if res.ok else None,
            **kwargs,
        )

    @classmethod
    def get_broadcasters_near_location(
        cls, latitude: float, longitude: float, meters: int, **kwargs
    ) -> List[models.RelativeBroadcasterLocation]:
        """Fetches a list of broadcasters near a location.

        :param latitude: The latitude of the search origin.
        :param longitude: The longitude of the search origin.
        :param meters: The distance from the origin to search.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A list of broadcaster relative location objects
        """
        update_kwargs_for_key(kwargs, "params", {"latitude": latitude, "longitude": longitude, "meters": meters})
        return cls.get(
            join_url_path(URL_PATH, "broadcasters", "near_location"),
            parse_func=lambda res: [models.RelativeBroadcasterLocation.parse(rec) for rec in parse_node(res).results],
            **kwargs,
        )

    @classmethod
    def get_sermon_event_types(cls, broadcaster_id: str = None, **kwargs) -> List[models.SermonEventTypeDetail]:
        """Fetches a list of valid sermon event types for a broadcaster.

        :param broadcaster_id: You can omit this if you are using a broadcaster API key.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A list of sermon event type detail objects
        """
        update_kwargs_for_key(kwargs, "params", {"broadcaster_id": broadcaster_id})
        return cls.get(
            join_url_path(URL_PATH, "filter_options", "sermon_event_types"),
            parse_func=lambda res: [models.SermonEventTypeDetail.parse(rec) for rec in parse_node(res).results],
            **kwargs,
        )

    #
    # Series
    #

    @classmethod
    def get_series_list(
        cls,
        broadcaster_id: str = None,
        page: int = 1,
        page_size: Optional[int] = None,
        sort_by: models.SeriesSortOrder = models.SeriesSortOrder.LAST_UPDATED,
        **kwargs,
    ) -> PaginatedResponse:  # noqa: F821
        """Fetches a sermon series list for a broadcaster.

        :param broadcaster_id: You can omit this if you are using a broadcaster API key.
        :param page: The page number to load (defaults to 1).
        :param page_size: The number of items per page (defaults to no limit).
        :param sort_by: A sort to apply to the results. Defaults to last updated.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A paginated response, with the results being SermonSeries objects.
        """
        update_kwargs_for_key(kwargs, "params", {"page": page, "pageSize": page_size, "sort_by": sort_by.value})

        path = join_url_path(URL_PATH, "broadcasters", broadcaster_id, "series")
        return cls.get_node(
            path,
            parse_func=lambda res: cls.PaginatedResponse(page, page_size, parse_node(res), models.SermonSeries.parse),
            **kwargs,
        )

    @classmethod
    def get_series(cls, series_name: str, broadcaster_id: str, **kwargs) -> Optional[models.SermonSeries]:
        """Fetches a single series.

        :param series_name: The name of the series.
        :param broadcaster_id: The broadcaster ID that owns the series.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A single models.SermonSeries object, or None if series name does not exist.
        """
        return cls.get(
            join_url_path(URL_PATH, "broadcasters", broadcaster_id, "series", series_name),
            parse_func=lambda res: models.SermonSeries.parse(res.json()) if res.ok else None,
            **kwargs,
        )

    #
    # Misc
    #

    @classmethod
    def get_speaker(cls, speaker_name: str, **kwargs) -> Optional[models.Speaker]:
        """Fetches a single speaker.

        :param speaker_name: The name of the speaker (exactly as it appears on the SA site).
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A single models.Speaker object, or None if speaker does not exist.
        """
        return cls.get(
            join_url_path(URL_PATH, "speakers", speaker_name),
            parse_func=lambda res: models.Speaker.parse(res.json()) if res.ok else None,
            **kwargs,
        )

    @classmethod
    def get_speakers(
        cls, broadcaster_id: Optional[str] = None, query: Optional[str] = None, page_size: Optional[int] = 25, **kwargs
    ) -> List[models.Speaker]:
        """Fetches a list of speakers.

        :param broadcaster_id: The ID of the broadcaster the speakers must have preached at.
        :param query: A search query to try to locate the speaker by.
        :param page_size: The number of results to return.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A list of models.Speaker objects
        """
        update_kwargs_for_key(
            kwargs, "params", {"broadcasterID": broadcaster_id, "query": query, "pageSize": page_size}
        )
        return cls.get(
            join_url_path(URL_PATH, "speakers"),
            parse_func=lambda res: [models.Speaker.parse(rec) for rec in parse_node(res).results],
            **kwargs,
        )

    @classmethod
    def get_webcasts_in_progress(cls, broadcaster_id: str = None, **kwargs) -> List[models.Webcast]:
        """Fetches a list of webcasts that are currently in progress.

        :param broadcaster_id: You can omit this if you are using a broadcaster API key.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A list of webcasts that are live now
        """
        update_kwargs_for_key(kwargs, "params", {"broadcasterID": broadcaster_id})
        return cls.get(
            join_url_path(URL_PATH, "webcasts"),
            parse_func=lambda res: [models.Webcast.parse(rec) for rec in parse_node(res).results],
            **kwargs,
        )

    @classmethod
    def get_spurgeon_devotional(
        cls, kind: models.SpurgeonDevotionalType, month: int, day_of_month: int, **kwargs
    ) -> Optional[models.SpurgeonDevotional]:
        """Returns the Spurgeon devotional for a given month and day

        :param kind: The kind of devotional (one of the values in models.SpurgeonDevotionalType).
        :param month: The month component of the date.
        :param day_of_month: The day of month component of the date.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A devotional object or None
        """
        update_kwargs_for_key(kwargs, "params", {"type": kind.value, "month": month, "day_of_month": day_of_month})

        def parse_spurgeon(result):
            data = parse_node(result).results

            return models.SpurgeonDevotional.parse(data) if data else None

        return cls.get(join_url_path(URL_PATH, "spurgeon_devotional"), parse_func=parse_spurgeon, **kwargs)

    @classmethod
    def get_filter_options(cls, broadcaster_id: str, **kwargs) -> Optional[models.FilterOptions]:
        """Returns filter options for the given broadcaster

        :param broadcaster_id: The broadcaster ID you are interested in.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A filter options object containing available filters for the broadcaster or None
        """
        return cls.get(
            join_url_path(URL_PATH, "broadcasters", broadcaster_id, "filter_options"),
            parse_func=lambda res: models.FilterOptions.parse(res.json()) if res.ok else None,
            **kwargs,
        )

    @classmethod
    def get_all_languages(cls, **kwargs) -> List[models.Language]:
        """Returns a list of all supported sermon languages

        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: [models.Language]
        """
        return cls.get(
            join_url_path(URL_PATH, "languages"),
            parse_func=lambda res: [models.Language.parse(rec) for rec in parse_node(res).results],
            **kwargs,
        )

    @classmethod
    def get_all_sermon_event_types(cls, **kwargs) -> List[models.Language]:
        """Returns a list of all supported sermon event types

        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: [models.Language]
        """
        return cls.get(
            join_url_path(URL_PATH, "filter_options", "sermon_event_types"),
            parse_func=lambda res: [models.SermonEventTypeDetail.parse(rec) for rec in parse_node(res).results],
            **kwargs,
        )

    @classmethod
    def get_speakers_for_broadcaster(cls, broadcaster_id: str = None, **kwargs) -> List[models.Speaker]:
        """Fetches a list of all speakers for a broadcaster.

        In contrast to get_speakers(), which searches across multiple
        broadcasters, this returns a list of all speakers associated
        with the given broadcaster.

        :param broadcaster_id: The ID of the broadcaster.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A list of models.Speaker objects.
        """
        update_kwargs_for_key(kwargs, "params", {"broadcasterID": broadcaster_id})
        return cls.get(
            join_url_path(URL_PATH, "broadcasters", broadcaster_id, "speakers"),
            parse_func=lambda res: [models.Speaker.parse(rec) for rec in parse_node(res).results],
            **kwargs,
        )

    @classmethod
    def get_speakers_for_series(cls, broadcaster_id: str, series_name: str, **kwargs) -> List[models.Speaker]:
        """Fetches a list of all speakers for a broadcaster's series.

        :param broadcaster_id: The ID of the broadcaster.
        :param series_name: The name of the series.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A list of models.Speaker objects.
        """
        return cls.get(
            join_url_path(URL_PATH, "broadcasters", broadcaster_id, "series", series_name, "speakers"),
            parse_func=lambda res: [models.Speaker.parse(rec) for rec in parse_node(res).results],
            **kwargs,
        )

    @classmethod
    def get_speaker_sermon_counts_for_broadcaster(
        cls, broadcaster_id: str = None, **kwargs
    ) -> List[models.SermonCountForSpeaker]:
        """Fetches a list of sermon counts for each speaker for a broadcaster.

        This method returns a list of all speakers for a specified
        broadcaster and the number of sermons preached for that
        broadcaster only. This restricted count is useful for ranking
        speakers by the number of sermons they've preached for a
        specific broadcaster. The total number of sermons preached by a
        speaker can be high for some guest speakers, but these speakers
        may have preached only few sermons for any given broadcaster.

        :param broadcaster_id: The ID of the broadcaster.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A list of models.SpeakerSermonCount objects.
        """
        update_kwargs_for_key(kwargs, "params", {"broadcasterID": broadcaster_id})
        return cls.get(
            join_url_path(URL_PATH, "broadcasters", broadcaster_id, "speakers", "sermon_counts"),
            parse_func=lambda res: [models.SermonCountForSpeaker.parse(rec) for rec in parse_node(res).results],
            **kwargs,
        )

    @classmethod
    def get_highlighted_sermons(cls, broadcaster_id: str, **kwargs) -> models.HighlightedSermons:
        """Fetches highlighted sermons info for a broadcaster.

        Each broadcaster can set a highlighted sermon. See the
        HighlightedAudioInputTypes enum for more documentation on the
        audio types you can set. (A separate welcome video can also be
        set. It is part of the Broadcaster model object.) The
        highlighted sermon appears on the broadcaster's home page. Also,
        a short list of sermons appears on a broadcaster's home
        page. The sermons appear in the sort order returned in the
        HighlightedSermons model object. API users can use this setting
        to create a similar list by calling the get_sermons() endpoint.

        :param broadcaster_id: The ID of the broadcaster to fetch.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: models.HighlightedSermons
        """
        return cls.get(
            join_url_path(URL_PATH, "broadcasters", str(broadcaster_id), "highlighted_sermons"),
            parse_func=lambda res: models.HighlightedSermons.parse(res.json()) if res.ok else None,
            **kwargs,
        )

    #
    #  Webcast stats
    #

    @classmethod
    def get_webcast(cls, webcast_id: int, broadcaster_id: str = None, **kwargs) -> models.Webcast:
        """Returns a single webcast for a broadcaster.

        :param broadcaster_id: Get webcast for this broadcaster.
        :param webcast_id: The webcast to retrieve.

        :return: A single webcast.
        """
        return cls.get(
            join_url_path(URL_PATH, "broadcasters", str(broadcaster_id), "webcasts", str(webcast_id)),
            parse_func=lambda res: models.Webcast.parse(res.json()) if res.ok else None,
            **kwargs,
        )

    @classmethod
    def get_webcasts(
        cls,
        broadcaster_id: str = None,
        start_time: Optional[int] = None,
        end_time: Optional[int] = None,
        sort_by: models.ListSortOrder = models.ListSortOrder.REVERSE_DATE,
        page: int = 1,
        page_size: Optional[int] = None,
        minimum_duration_seconds: Optional[int] = None,
        **kwargs,
    ) -> PaginatedResponse:  # noqa: F821
        """Returns a list of webcasts for a broadcaster.

        While get_webcasts_in_progress() returns only webcasts which are
        currently live, get_webcasts() returns information about all
        webcasts. get_stats_for_webcast() can be used to retrieve
        detailed information about a specific webcast.

        If a date range is supplied, the list of broadcasts is
        restricted to webcasts in that date range. Otherwise, all
        webcasts are returned. Because a webcast can span more than one
        day, only the start time of a webcast is used to create the
        range.

        :param broadcaster_id: Get webcasts for this broadcaster.
        :param start_time: Optional start of date range as UNIX timestamp in seconds.
        :param end_time: Optional end of date range as UNIX timestamp in seconds.
        :param page: The page number to load (defaults to 1).
        :param page_size: The number of items per page.
        :param minimum_duration_seconds: If present, webcasts must have at least
        this duration. Useful for filtering out short test webcasts which can
        distort stats.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: A list of webcasts for the broadcaster
        """
        update_kwargs_for_key(
            kwargs,
            "params",
            {
                "sortBy": sort_by.value,
                "startTime": start_time,
                "endTime": end_time,
                "page": page,
                "pageSize": page_size,
                "minimumDurationSeconds": minimum_duration_seconds,
            },
        )

        return cls.get_node(
            join_url_path(URL_PATH, "broadcasters", broadcaster_id, "webcasts"),
            parse_func=lambda res: cls.PaginatedResponse(page, page_size, parse_node(res), models.Webcast.parse),
            **kwargs,
        )

    @classmethod
    def get_webcast_stats(cls, webcast_id: int, **kwargs) -> models.WebcastStats:
        """Fetches stats for a given webcast.

        :param webcast_id: The webcast
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: models.WebcastStats
        """

        return cls.get(
            join_url_path("webcasts", str(webcast_id), "stats"),
            parse_func=lambda res: models.WebcastStats.parse(res.json()) if res.ok else None,
            **kwargs,
        )

    @classmethod
    def get_webcast_stats_totals(
        cls, start_time: int, end_time: int, broadcaster_id: str = None, **kwargs
    ) -> models.WebcastRangeTotals:
        """Fetches total stats over a date range for webcasts.

        :param start_time: Start of date range as UNIX timestamp in seconds.
        :param end_time: End of date range as UNIX timestamp in seconds.
        :param kwargs: Additional arguments to pass to the underlying get method.
        :return: models.WebcastRangeTotals
        """

        update_kwargs_for_key(kwargs, "params", {"startTime": start_time, "endTime": end_time})

        return cls.get(
            join_url_path(URL_PATH, "broadcasters", str(broadcaster_id), "webcasts", "stats", "totals"),
            parse_func=lambda res: models.WebcastRangeTotals.parse(res.json()) if res.ok else None,
            **kwargs,
        )



import functools
import ssl
import locale

from http import HTTPStatus
from typing import Optional
from urllib.parse import urljoin

from requests import Session
from requests.adapters import HTTPAdapter
from urllib3 import Retry

_api_key = None
_base_url = ""
_preferred_language = ""
_session = Session()


class SecureRequestAdapter(HTTPAdapter):
    def init_poolmanager(self, connections, maxsize, block=False, **kwargs):
        kwargs = kwargs or {}
        kwargs["ssl_version"] = ssl.PROTOCOL_TLSv1_2

        super().init_poolmanager(connections, maxsize, block, **kwargs)


# Require TLSv1.2 for all connections to the server
_session.mount("https://", SecureRequestAdapter())


def set_base_url(url: str):
    global _base_url
    _base_url = url


def get_base_url() -> str:  # pragma: no cover
    return _base_url


def set_api_key(key: str):
    """Sets the SermonAudio API key to use for authenticated requests. A key must be set before making any requests.

    :param key: Your API key. Broadcasters may obtain an API key by going to the Member's Only area of the site. If you
    are an application developer and would like to integrate SermonAudio into your product or website, please contact
    us at info@sermonaudio.com and we'll be happy to discuss with you.
    """
    global _api_key
    _api_key = key


def set_preferred_language(http_accept_language: str):
    """Sets the preferred language string globally. Defaults to your system locale environment variables.

    The format of the string should be an acceptable HTTP Accept-Language header. You may specify multiple languages,
    and the API will attempt to accommodate your preferences when possible. While content language will always be
    determined by the broadcasters individually, other strings, whenever possible, will be translated."""
    global _preferred_language
    _preferred_language = http_accept_language


# Set reasonable defaults
set_base_url("https://api.sermonaudio.com/v2/")
set_preferred_language((locale.getdefaultlocale()[0] or "en-US").replace("_", "-"))


def create_or_configure_session_with_retry(
    retries=3, backoff_factor=0.3, status_forcelist=(500, 502, 504), session=None
) -> Session:
    """Creates or configures a Session with retry logic.

    Only use this if you know what you're doing. You should be VERY careful about using this
    kind of logic for any non-GET requests. You can pass the session that you create (please re-use it!)
    to any request methods you call that need retries via the keyword argument `session`.
    """
    session = session or Session()
    retry = Retry(
        total=retries, read=retries, connect=retries, backoff_factor=backoff_factor, status_forcelist=status_forcelist
    )
    adapter = SecureRequestAdapter(max_retries=retry)

    session.mount("http://", adapter)
    session.mount("https://", adapter)

    return session


class API:
    @classmethod
    def _get_request_headers(cls, preferred_language_override=None, show_content_in_any_language=True):
        result = {"X-API-Key": _api_key, "Accept-Language": preferred_language_override or _preferred_language}

        if show_content_in_any_language:
            result["X-Show-All-Languages"] = "True"

        return result

    @classmethod
    def _request(
        cls,
        path: str,
        method: str = "get",
        headers: Optional[dict] = None,
        preferred_language_override=None,
        show_content_in_any_language=True,
        session: Session = _session,
        parse_func=None,
        api_key_override: str = None,
        **kwargs,
    ):
        """Low-level API call helper.

        :param method: The HTTP method of the request.
        :param path: The path to the endpoint you want to call (e.g. /node/sermons/1234).
        :param headers: Additional headers to pass (the library handles authentication etc. automatically).
        :param preferred_language_override: An optional language preference override. The system default will be used.
        :param show_content_in_any_language: When false, only shows content in the preferred langauge.
        :param session: The request.Sesion object to use when making the request. Don't use without a good reason.
        :param parse_func: If present, this function will be invoked with the response, and the result of that
        invocation will be returned instead.
        :param kwargs: Any additional keyword arguments will be passed directly to session.request(). Note that in
        general, the params argument, when present, should be respected by all downstream methods. They will generally
        do a dict.update() of any params that they require, so if you add something stupid, expect it to be overridden.
        Otherwise, feel free to pass additional params. This allows for clean extension for internal use.
        """
        # Generate the URL using the path provided and the base URL
        url = urljoin(_base_url, path)

        # Update the headers passed in with our own required ones
        headers = headers or {}
        base_headers = cls._get_request_headers(preferred_language_override, show_content_in_any_language)
        headers.update(base_headers)

        if api_key_override:
            headers["X-API-Key"] = api_key_override  # pragma: no cover

        response = session.request(method, url, headers=headers, **kwargs)

        return parse_func(response) if parse_func is not None else response

    get = functools.partialmethod(_request, method="get")
    post = functools.partialmethod(_request, method="post")
    put = functools.partialmethod(
        _request,
        method="put",
        parse_func=lambda res: res.status_code in {HTTPStatus.OK, HTTPStatus.CREATED, HTTPStatus.NO_CONTENT},
    )
    patch = functools.partialmethod(
        _request,
        method="patch",
        parse_func=lambda res: res.status_code in {HTTPStatus.OK, HTTPStatus.ACCEPTED, HTTPStatus.NO_CONTENT},
    )
    delete = functools.partialmethod(
        _request,
        method="delete",
        parse_func=lambda res: res.status_code in {HTTPStatus.OK, HTTPStatus.ACCEPTED, HTTPStatus.NO_CONTENT},
    )


class APIException(Exception):
    pass



from setuptools import setup

with open("README.md") as fp:
    readme = fp.read()

setup(
    name="sermonaudio",
    version="6.6.0",
    description="The official Python client library for accessing the SermonAudio.com APIs",
    long_description=readme,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
    ],
    url="http://api.sermonaudio.com/",
    author="SermonAudio.com",
    author_email="info@sermonaudio.com",
    keywords="sermon audio sermonaudio API preaching church bible",
    license="MIT",
    packages=["sermonaudio", "sermonaudio.node", "sermonaudio.broadcaster"],
    install_requires=["requests>=2.19.1", "pytz>=2018.5", "osis-book-tools>=1.0.1"],
)
